from neural2 import *
from activations import *
import colors
from layers import *
import iterators

__author__ =    "Jonathan N Lee"
__license__ =   "MIT"
__version__ =   "1.2.0"
__email__ =     "jonathan_lee@berkeley.edu"