# 
# @author jonathan_lee@berkeley.edu (Jonathan N. Lee)
# 
# Don't bother reading this. These are just color codes.
#

header = '\033[95m'
blue = '\033[94m'
green = '\033[92m'
yellow = '\033[93m'
red = '\033[91m'
end = '\033[0m'
bold = '\033[1m'
underline = '\033[4m'
